@@ -1,8 +1,8 @@
-;;; vc-fossil.el --- VC backend for the fossil sofware configuraiton management system
+;;; vc-fossil.el --- VC backend for the fossil sofware configuraiton management system  -*- lexical-binding: t; -*-
 
 ;; Author: Venkat Iyer <venkat@comit.com>
 ;; Maintainer: Alfred M. Szmidt <ams@gnu.org>
-;; Version: 20210928
+;; Version: 20220707
 
 ;; vc-fossil.el free software: you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published
@@ -51,12 +51,22 @@
 ;; * checkout (file &optional rev)		OK
 ;; * revert (file &optional contents-done)	OK
 ;; - merge-file (file &optional rev1 rev2)	??
-;; - merge-branch ()				??
+;; - merge-branch ()				OK
 ;; - merge-news (file)				??
 ;; - pull (prompt)				OK
 ;; ? push (prompt)				OK
 ;; - steal-lock (file &optional revision)	??
-;; - modify-change-comment (files rev comment)	??
+;; - modify-change-comment (files rev comment)	BROKEN
+;;     This requires a different version of LOG-VIEW-EXTRACT-COMMENT
+;;     and LOG-VIEW-CURRENT-FILE to work.
+;;
+;;     For LOG-VIEW-CURRENT-FILE there has been a bug report filed
+;;     with a fix for GNU Emacs
+;;     (https://lists.gnu.org/archive/html/emacs-devel/2022-05/msg00759.html).
+;;
+;;     LOG-VIEW-EXTRACT-COMMENT needs to be fixed as well somehow to
+;;     extract the actual log message around point.
+;;
 ;; - mark-resolved (files)			??
 ;; - find-admin-dir (file)			??
 ;; HISTORY FUNCTIONS
@@ -100,8 +110,8 @@
 
 ;;; Code:
 
-(eval-when-compile
-  (require 'vc))
+(eval-when-compile (require 'vc))
+(eval-when-compile (require 'url-parse)) ;For url-user/passwd setters
 
 (autoload 'vc-switches "vc")
 
@@ -150,7 +160,7 @@
       (setq fossil-program (car args)
 	    command (cadr args)
 	    args (cddr args)))
-    (apply 'vc-do-async-command buffer root fossil-program command args)
+    (apply #'vc-do-async-command buffer root fossil-program command args)
     (with-current-buffer buffer
       (vc-run-delayed (vc-compilation-mode 'Fossil)))
     (vc-set-async-update buffer)))
@@ -158,7 +168,7 @@
 (defun vc-fossil--get-id (dir)
   (let* ((default-directory dir)
 	 (info (vc-fossil--run "info"))
-	 (pos (string-match "checkout: *\\([0-9a-fA-F]+\\)" info))
+	 (_pos (string-match "checkout: *\\([0-9a-fA-F]+\\)" info))
 	 (uid (match-string 1 info)))
     (substring uid 0 10)))
 
@@ -201,6 +211,18 @@
     (dolist (l (split-string (vc-fossil--run "remote" "list") "\n" t))
       (push (split-string l) remotes))
     remotes))
+
+(defun vc-fossil--branches ()
+  "Return the existing branches, as a list of strings.
+The car of the list is the current branch."
+  (with-temp-buffer
+    ;;;---!!! This requires that fossil is compiled with JSON support.
+    (vc-fossil--call t "json" "branch" "list")
+    (goto-char (point-min))
+    (let* ((payload (gethash "payload" (json-parse-buffer)))
+	   (current-branch (gethash "current" payload))
+	   (branches (append (gethash "branches" payload) nil)))
+      (cons current-branch (remove current-branch branches)))))
 
 ;; Customization
 
@@ -214,14 +236,12 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
   :type '(choice (const :tag "Unspecified" nil)
 		 (const :tag "None" t)
 		 (string :tag "Argument String")
-		 (repeat :tag "Argument List" :value ("") string))
-  :group 'vc-fossil)
+		 (repeat :tag "Argument List" :value ("") string)))
 
 (defcustom vc-fossil-extra-header-fields (list :repository :remote-url :checkout :tags)
   "A list of keywords denoting extra header fields to show in the vc-dir buffer."
   :type '(set (const :repository) (const :remote-url) (const :synchro)
-	      (const :checkout) (const :comment) (const :tags))
-  :group 'vc-fossil)
+	      (const :checkout) (const :comment) (const :tags)))
 
 ;; BACKEND PROPERTIES
 
@@ -231,7 +251,13 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 
 ;; STATE-QUERYING FUNCTIONS
 
-;;;###autoload
+
+;;;###autoload (defun vc-fossil-registered (file)
+;;;###autoload   "Return non-nil if FILE is registered with Fossil."
+;;;###autoload   (if (vc-find-root file ".fslckout")       ; Short cut.
+;;;###autoload       (progn
+;;;###autoload         (load "vc-fossil" nil t)
+;;;###autoload         (vc-fossil-registered file))))
 (defun vc-fossil-registered (file)
   (with-temp-buffer
     (let* ((str (ignore-errors
@@ -257,10 +283,9 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 
 (defun vc-fossil-dir-status-files (dir files update-function)
   (vc-fossil--classify-all-files dir)
-  (insert (apply 'vc-fossil--run "changes" "--classify" "--all"
+  (insert (apply #'vc-fossil--run "changes" "--classify" "--all"
 		 (or files (list dir))))
-  (let ((result '())
-	(root (vc-fossil-root dir)))
+  (let ((result '()))
     (goto-char (point-min))
     (while (not (eobp))
       (let* ((line (buffer-substring-no-properties (point) (line-end-position)))
@@ -272,13 +297,14 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 	    ;; If 'fossil update' says file is UNCHANGED check to see
 	    ;; if it has been RENAMED.
 	    (when (or (not state) (eql state 'up-to-date))
-	      (setq state (vc-fossil--state-code (cdr (assoc file vc-fossil--file-classifications)))))
+	      (setq state (vc-fossil--state-code
+			   (cdr (assoc file vc-fossil--file-classifications)))))
 	    (when (not (eq state 'up-to-date))
 	      (push (list file state) result))))
 	(forward-line)))
     ;; Now collect untracked files.
     (delete-region (point-min) (point-max))
-    (insert (apply 'vc-fossil--run "extras" "--dotfiles" (list dir)))
+    (insert (apply #'vc-fossil--run "extras" "--dotfiles" (list dir)))
     (goto-char (point-min))
     (while (not (eobp))
       (let ((file (buffer-substring-no-properties (point) (line-end-position))))
@@ -286,7 +312,7 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 	(forward-line)))
     (funcall update-function result nil)))
 
-(defun vc-fossil-dir-extra-headers (dir)
+(defun vc-fossil-dir-extra-headers (_dir)
   (let ((info (vc-fossil--run "info"))
 	(settings (vc-fossil--run "settings"))
 	(lines nil))
@@ -306,12 +332,13 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 		    (autosync (and as-match (match-string 1 settings)))
 		    (dp-match (string-match "^dont-push +.+ +\\([[:graph:]]+\\)$" settings))
 		    (dontpush (and dp-match (match-string 1 settings))))
-	       (push (vc-fossil--propertize-header-line "Synchro    : "
-							(concat (and autosync "autosync=") autosync
-								(and dontpush " dont-push=") dontpush))
+	       (push (vc-fossil--propertize-header-line
+		      "Synchro    : "
+		      (concat (and autosync "autosync=") autosync
+			      (and dontpush " dont-push=") dontpush))
 		     lines)))
 	    ((eql field :checkout)
-	     (let* ((posco (string-match "checkout: *\\([0-9a-fA-F]+\\) \\([-0-9: ]+ UTC\\)" info))
+	     (let* ((_posco (string-match "checkout: *\\([0-9a-fA-F]+\\) \\([-0-9: ]+ UTC\\)" info))
 		    (coid (substring (match-string 1 info) 0 10))
 		    (cots (format-time-string "%Y-%m-%d %H:%M:%S %Z"
 					      (safe-date-to-time (match-string 2 info))))
@@ -340,7 +367,7 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
     (and line
 	 (cadr (split-string line)))))
 
-(defun vc-fossil-checkout-model (files) 'implicit)
+(defun vc-fossil-checkout-model (_files) 'implicit)
 
 ;; - mode-line-string (file)
 
@@ -349,7 +376,7 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 (defun vc-fossil-create-repo ()
   (vc-fossil--command nil 0 nil "new"))
 
-(defun vc-fossil-register (files &optional rev comment)
+(defun vc-fossil-register (files &optional _rev _comment)
   ;; We ignore the comment.  There's no comment on add.
   (vc-fossil--command nil 0 files "add"))
 
@@ -361,8 +388,8 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 (defun vc-fossil-unregister (file)
   (vc-fossil--command nil 0 file "rm"))
 
-(defun vc-fossil-checkin (files comment &optional rev)
-  (apply 'vc-fossil--command nil 0 files
+(defun vc-fossil-checkin (files comment &optional _rev)
+  (apply #'vc-fossil--command nil 0 files
 	 (nconc (list "commit" "-m")
 		(log-edit-extract-headers
 		 `(("Author" . "--user-override")
@@ -377,7 +404,7 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 	  (unless (zerop (length rev)) (list "-r" rev))
 	  (vc-switches 'Fossil 'checkout))))
 
-(defun vc-fossil-checkout (file &optional editable rev)
+(defun vc-fossil-checkout (file &optional _editable rev)
   (apply #'vc-fossil--command nil 0 file
 	 "update"
 	 (nconc
@@ -394,7 +421,16 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 
 ;; - merge-file (file &optional rev1 rev2)
 
-;; - merge-branch ()
+(defun vc-fossil-merge-branch ()
+  "Merge changes into the current branch.
+This prompts for a branch to merge from."
+  (let* ((root (vc-fossil-root default-directory))
+	 (buffer (format "*vc-fossil : %s*" (expand-file-name root)))
+	 (branches (cdr (vc-fossil--branches)))
+	 (merge-source (completing-read "Merge from branch: " branches nil t)))
+    (apply #'vc-do-async-command buffer root "fossil" "merge" (list merge-source))
+    (with-current-buffer buffer (vc-run-delayed (vc-compilation-mode 'Fossil)))
+    (vc-set-async-update buffer)))
 
 ;; - merge-news (file)
 
@@ -411,7 +447,8 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 
 ;; - steal-lock (file &optional revision)
 
-;; - modify-change-comment (files rev comment)
+(defun vc-fossil-modify-change-comment (_files rev comment)
+  (vc-fossil--call t "amend" rev "-m" comment))
 
 ;; - mark-resolved (files)
 
@@ -419,7 +456,7 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 
 ;; HISTORY FUNCTIONS
 
-(defun vc-fossil-print-log (files buffer &optional shortlog start-revision limit)
+(defun vc-fossil-print-log (files buffer &optional _shortlog start-revision limit)
   ;; TODO: We actually already have short, start and limit, need to
   ;; add it into the code.
   (vc-setup-buffer buffer)
@@ -430,7 +467,8 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 	       (nconc
 		(when start-revision (list "before" start-revision))
 		(when limit (list "-n" (number-to-string limit)))
-		(list "-p" (file-relative-name (expand-file-name file)))))))))
+		(list "-p" (file-relative-name (expand-file-name file))))))
+      (goto-char (point-min)))))
 
 ;; * log-outgoing (buffer remote-location)
 
@@ -445,23 +483,23 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 
 (define-derived-mode vc-fossil-log-view-mode log-view-mode "Fossil-Log-View"
   (setq word-wrap t)
-  (set (make-local-variable 'wrap-prefix) "                      ")
-  (set (make-local-variable 'log-view-file-re) "\\`a\\`")
-  (set (make-local-variable 'log-view-per-file-logs) nil)
-  (set (make-local-variable 'log-view-message-re)
-       "^[0-9:]+ \\[\\([0-9a-fA-F]*\\)\\] \\(?:\\*[^*]*\\*\\)? ?\\(.*\\)")
-  (set (make-local-variable 'log-view-font-lock-keywords)
-       (append
-	'(
-	  ("^\\([0-9:]*\\) \\(\\[[[:alnum:]]*\\]\\) \\(\\(?:\\*[[:word:]]*\\*\\)?\\) ?\\(.*?\\) (user: \\([[:word:]]*\\) tags: \\(.*\\))"
-	   (1 'change-log-date)
-	   (2 'change-log-name)
-	   (3 'highlight)
-	   (4 'log-view-message)
-	   (5 'change-log-name)
-	   (6 'highlight))
-	  ("^=== \\(.*\\) ==="
-	   (1 'change-log-date))))))
+  (setq-local wrap-prefix "                      ")
+  (setq-local log-view-file-re "\\`a\\`")
+  (setq-local log-view-per-file-logs nil)
+  (setq-local log-view-message-re
+	      "^[0-9:]+ \\[\\([0-9a-fA-F]*\\)\\] \\(?:\\*[^*]*\\*\\)? ?\\(.*\\)")
+  (setq-local log-view-font-lock-keywords
+	      (append
+	       '(
+		 ("^\\([0-9:]*\\) \\(\\[[[:alnum:]]*\\]\\) \\(\\(?:\\*[[:word:]]*\\*\\)?\\) ?\\(.*?\\) (user: \\([[:word:]]*\\) tags: \\(.*\\))"
+		  (1 'change-log-date)
+		  (2 'change-log-name)
+		  (3 'highlight)
+		  (4 'log-view-message)
+		  (5 'change-log-name)
+		  (6 'highlight))
+		 ("^=== \\(.*\\) ==="
+		  (1 'change-log-date))))))
 
 ;; - show-log-entry (revision)
 
@@ -469,7 +507,7 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 
 ;; - update-changelog (files)
 
-(defun vc-fossil-diff (files &optional rev1 rev2 buffer async)
+(defun vc-fossil-diff (files &optional rev1 rev2 buffer _async)
   ;; TODO: Implement diff for directories.
   (let ((buf (or buffer "*vc-diff*"))
 	(root (and files (expand-file-name (vc-fossil-root (car files))))))
@@ -492,7 +530,7 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
   "\\([[:word:]]+\\)\\s-+\\([-0-9]+\\)\\s-+[0-9]+: ")
 
 (defun vc-fossil-annotate-command (file buffer &optional rev)
-  (vc-fossil--command buffer 0 file "annotate"))
+  (vc-fossil--command buffer 0 file "annotate" "-r" (or rev "trunk")))
 
 (defun vc-fossil-annotate-time ()
   ;; TODO: Currently only the date is used, not the time.
@@ -526,7 +564,7 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 						'("tag" "add"))
 					    ,name ,(vc-fossil--get-id dir)))))
 
-(defun vc-fossil-retrieve-tag (dir name update)
+(defun vc-fossil-retrieve-tag (dir name _update)
   (let ((default-directory dir))
     (vc-fossil--command nil 0 nil "checkout" name)))
 
@@ -615,30 +653,31 @@ If nil, use the value of `vc-diff-switches'.  If t, use no switches."
 	   (setq file (match-string 1 l0)))
       file)))
 
-(defun vc-fossil-link (start end)
-  "Puts the current URL to a file in the kill ring."
+(defun vc-fossil-link (_start _end)
+  "Put the current URL to a file in the kill ring."
   (interactive "r")
   (let ((default-directory (file-name-directory (buffer-file-name (current-buffer)))))
     (unless (vc-fossil-registered (buffer-file-name))
       (error "%s: file is not registerd in vc" (buffer-file-name)))
     (let* ((repository-url (vc-fossil--url-without-authinfo
 			    (vc-fossil-repository-url (buffer-file-name))))
-   	   (file (vc-fossil--relative-file-name (buffer-file-name)))
+	   (file (vc-fossil--relative-file-name (buffer-file-name)))
 	   (tag (vc-fossil-working-revision (buffer-file-name (current-buffer))))
    	   (start (line-number-at-pos (region-beginning)))
-   	   (end (line-number-at-pos (region-end))))
-      (if (= start end)
-	  (setq link (format "%s/file?ci=%s&name=%s&ln=%s"
- 			     repository-url tag file start))
-	(setq link (format "%s/file?ci=%s&name=%s&ln=%s-%s"
- 			   repository-url tag file start end)))
+   	   (end (line-number-at-pos (region-end)))
+           (link
+            (if (= start end)
+                (format "%s/file?ci=%s&name=%s&ln=%s"
+ 			repository-url tag file start)
+              (format "%s/file?ci=%s&name=%s&ln=%s-%s"
+ 		      repository-url tag file start end))))
       (kill-new link)
       (message "%s" link))))
 
-;;; This snippet enables the Fossil VC backend so it will work once
-;;; this file is loaded.  By also marking it for inclusion in the
-;;; autoloads file, installing packaged versions of this should work
-;;; without users having to monkey with their init files.
+;; This snippet enables the Fossil VC backend so it will work once
+;; this file is loaded.  By also marking it for inclusion in the
+;; autoloads file, installing packaged versions of this should work
+;; without users having to monkey with their init files.
 
 ;;;###autoload
 (add-to-list 'vc-handled-backends 'Fossil t)
